@@ -43,25 +43,77 @@
 # those contributions and any derivatives thereof.
 # 
 # END BPS TAGGED BLOCK }}}
+use 5.008003;
 use strict;
 use warnings; no warnings qw(redefine);
 
 package RT::Extension::MergeUsers;
 
-our $VERSION = '0.04_01';
+our $VERSION = '0.12_01';
+
+=head1 NAME
+
+RT::Extension::MergeUsers - Merges two users into the same effective user
+
+=head1 DESCRIPTION
+
+This RT extension adds a "Merge Users" box to the User Administration page, 
+which allows you to merge the user you are currently viewing with another
+user on your RT instance.
+
+It also adds L</MergeInto> and L</UnMerge> functions to the L<RT::User> class,
+which allow you to programmatically accomplish the same thing from your code.
+
+It also provides a version of L<CanonicalizeEmailAddress>, which means that
+all e-mail sent from secondary users is displayed as coming from the primary
+user.
+
+=head1 INSTALLATION
+
+If you're upgrading then, as well, read L</UPGRADING> below.
+
+    perl Makefile.PL
+    make
+    make install
+    clear your mason cache
+        most often this would be rm -rf /opt/rt3/var/mason_data/*
+    restart apache
+
+For RT 3.8
+
+    Add RT::Extension::MergeUsers to your /opt/rt3/etc/RT_SiteConfig.pm file
+    Set(@Plugins, qw(RT::Extension::MergeUsers));
+
+    If you have more than one Plugin enabled, you must enable them as one 
+    Set(@Plugins, qw(Foo Bar)); command
+
+=head1 UPGRADING
+
+If you are upgrading from 0.03_01 or earlier, you must run F<rt-upgrade-merged-users>.
+This script will create MergedUsers Attributes so RT can know when you're looking
+at a user that other users have been merged into. If you don't run this script,
+you'll have issues unmerging users. It can be safely run multiple times, it will
+only create Attributes as needed.
+
+=cut
 
 package RT::User;
 
 our %EFFECTIVE_ID_CACHE;
 
 use RT::Interface::Web::Handler;
-use Hook::LexWrap;
 
-{ my $i = 0;
-wrap 'RT::Interface::Web::Handler::CleanupRequest', post => sub {
-    return if ++$i%100; # flush cache every N requests
-    %EFFECTIVE_ID_CACHE = ();
-}; }
+{
+    my $i = 0;
+
+    my $old_cleanup = \&RT::Interface::Web::Handler::CleanupRequest;
+    no warnings 'redefine';
+    *RT::Interface::Web::Handler::CleanupRequest = sub {
+        $old_cleanup->(@_);
+        return if ++$i % 100; # flush cache every N requests
+        %EFFECTIVE_ID_CACHE = ();
+    };
+}
 
 sub CanonicalizeEmailAddress {
     my $self = shift;
@@ -72,18 +124,10 @@ sub CanonicalizeEmailAddress {
     }
 
     # get the user whose email address this is
-    my $canonical_user = RT::User->new($RT::SystemUser);
+    my $canonical_user = RT::User->new( $self->CurrentUser );
     $canonical_user->LoadByCols( EmailAddress => $address );
     return $address unless $canonical_user->id;
-
-    # if we got a user, check for a parent
-    my ($effective_id) = $canonical_user->Attributes->Named("EffectiveId");
-    return $address unless $effective_id && $effective_id->id;
-
-    $canonical_user->LoadByCols( id => $effective_id->Content );
-    return $address unless $canonical_user->id;
-
-    # is there another parent user above this one?
+    return $address unless $canonical_user->EmailAddress ne $address;
     return $canonical_user->CanonicalizeEmailAddress(
         $canonical_user->EmailAddress
     );
@@ -98,7 +142,7 @@ sub LoadByCols {
         my $effective_id = RT::Attribute->new( $RT::SystemUser );
         $effective_id->LoadByCols(
             Name       => 'EffectiveId',
-            ObjectType => ref($self) || $self,
+            ObjectType => __PACKAGE__,
             ObjectId   => $oid,
         );
         if ( $effective_id->id && $effective_id->Content && $effective_id->Content != $oid ) {
@@ -143,30 +187,22 @@ sub MergeInto {
     if (ref $user) {
         return (0, "User is not loaded") unless $user->id;
 
-        $merge = RT::User->new($RT::SystemUser);
+        $merge = RT::User->new( $self->CurrentUser );
         $merge->Load($user->id);
         return (0, "Could not reload user #". $user->id)
             unless $merge->id;
     } else {
-        $merge = RT::User->new($RT::SystemUser);
+        $merge = RT::User->new( $self->CurrentUser );
         $merge->Load($user);
         return (0, "Could not load user '$user'") unless $merge->id;
     }
+    return (0, "Could not load user to be merged") unless $merge->id;
 
     # Get copies of the canonicalized users
     my $email;
-    if (defined $merge->Attributes->Named('EffectiveId')) {
-        $email = $merge->CanonicalizeEmailAddress($merge->EmailAddress);
-        $merge->LoadByEmail($email);
-    }
-    return (0, "Could not load user to be merged") unless $merge->id;
 
-    my $canonical_self = RT::User->new($RT::SystemUser);
+    my $canonical_self = RT::User->new( $self->CurrentUser );
     $canonical_self->Load($self->id);
-    if (defined $canonical_self->Attributes->Named('EffectiveId')) {
-        $email = $canonical_self->CanonicalizeEmailAddress($canonical_self->EmailAddress);
-        $canonical_self->LoadByEmail($email);
-    }
     return (0, "Could not load user to merge into") unless $canonical_self->id;
 
     # No merging into yourself!
@@ -179,6 +215,9 @@ sub MergeInto {
     return (0, "User @{[$canonical_self->Name]} has already been merged")
            if defined $new and $new->Content == $canonical_self->id;
 
+    # clean the cache
+    delete $EFFECTIVE_ID_CACHE{$self->id};
+
     # do the merge
     $canonical_self->SetAttribute(
         Name => "EffectiveId",
@@ -189,13 +228,14 @@ sub MergeInto {
     my $merged_users = $merge->GetMergedUsers;
     $merged_users->SetContent( [$canonical_self->Id, @{$merged_users->Content}] );
 
-    $canonical_self->SetComments( join "\n", grep /\S/,
-        $canonical_self->Comments||'',
-        "Merged into ". $merge->EmailAddress ." (". $merge->id .")",
-    );
     $merge->SetComments(join "\n", grep /\S/,
         $merge->Comments||'',
-        $canonical_self->EmailAddress." (".$canonical_self->id.") merged into this user",
+        ($canonical_self->EmailAddress || $canonical_self->Name)." (".$canonical_self->id.") merged into this user",
+    );
+
+    $canonical_self->SetComments( join "\n", grep /\S/,
+        $canonical_self->Comments||'',
+        "Merged into ". ($merge->EmailAddress || $merge->Name)." (". $merge->id .")",
     );
     return (1, "Merged users successfuly");
 }
@@ -208,20 +248,20 @@ sub UnMerge {
 
     # flush the cache, or the Sets below will
     # clobber $self
-    delete $EFFECTIVE_ID_CACHE{$self->Id};
+    delete $EFFECTIVE_ID_CACHE{$self->id};
 
-    my $merge = RT::User->new($RT::SystemUser);
+    my $merge = RT::User->new( $self->CurrentUser );
     $merge->Load( $current->Content );
 
     $current->Delete;
     $self->SetComments( join "\n", grep /\S/,
         $self->Comments||'',
-        "Unmerged from ". $merge->EmailAddress ." (".$merge->id.")",
+        "Unmerged from ". ($merge->EmailAddress || $merge->Name) ." (".$merge->id.")",
     );
 
     $merge->SetComments(join "\n", grep /\S/,
         $merge->Comments,
-        $self->EmailAddress ." (". $self->id .") unmerged from this user",
+        ($self->EmailAddress || $self->Name) ." (". $self->id .") unmerged from this user",
     );
 
     my $merged_users = $merge->GetMergedUsers;
@@ -239,16 +279,16 @@ sub SetEmailAddress {
     my $self = shift;
     my $value = shift;
 
-    return ( 0, $self->loc('Email address in use') )
-        unless $self->ValidateEmailAddress( $value );
+    my ( $val, $msg ) = $self->ValidateEmailAddress($value);
+    return ( 0, $msg || $self->loc('Email address in use') ) unless $val;
 
     # if value is valid then either there is no user or
     # user is merged into this one
-    my $tmp = RT::User->new($RT::SystemUser);
+    my $tmp = RT::User->new( $self->CurrentUser );
     $tmp->LoadOriginal( EmailAddress => $value );
     if ( $tmp->id && $tmp->id != $self->id ) {
         # there is a different user record
-        $tmp->_Set( Field => 'EmailAddress', Value => undef );
+        $tmp->_Set( Field => 'EmailAddress', Value => "" );
     }
 
     return $self->_Set( Field => 'EmailAddress', Value => $value );
@@ -268,7 +308,7 @@ sub NameAndEmail {
 
 package RT::Users;
 use RT::Users;
-        
+
 =head2 Next
 
 This custom iterator makes sure that duplicate users are never shown in search results.
@@ -279,8 +319,7 @@ sub Next {
     my $self = shift;
     
     my $user = $self->SUPER::Next(@_);
-    return unless ($user && $user->id);
-    unless ($user) {
+    unless ($user and $user->id) {
         $self->{seen_users} = undef;
         return undef;
     }   
@@ -294,7 +333,22 @@ sub Next {
     return $self->Next() if ($self->{seen_users}->{$user->id}++);
 
     return $user;
-
 }
 
+sub GotoFirstItem {
+    my $self = shift;
+    $self->{seen_users} = undef;
+    $self->GotoItem(0);
+}
+
+=head1 AUTHOR
+
+Alex Vandiver E<lt>alexmv@bestpractical.comE<gt>
+
+=head1 LICENSE
+
+GPL version 2.
+
+=cut
+
 1;
